import sys

from names_matcher.algorithm import NamesMatcher


def main():
    """
    Given two text files with one identity per line, output the matches from \
    the first to the second.

    The names are expected to be joined by "|". The result is printed to stdout, lines from
    the second file in the sequence of the first file.
    """
    if len(sys.argv) != 3:
        print("Usage: path/to/file/1 path/to/file/2", file=sys.stderr)
        return 1
    names = [None] * 2
    for arg in (1, 2):
        with open(sys.argv[arg]) as fin:
            names[arg - 1] = [line.split("|") for line in fin]
    import logging
    log = logging.getLogger()
    log.info("read")
    for match in NamesMatcher()(*names)[0]:
        if match < 0:
            print()
        else:
            sys.stdout.write("|".join(names[1][match]))


if __name__ == "__main__":
    sys.exit(main())



import os
import platform
import sys
from tempfile import NamedTemporaryFile
from unittest.mock import patch

from names_matcher.__main__ import main


def test_main(capsys):
    is_windows = platform.system() == "Windows"
    with NamedTemporaryFile(mode="w+", prefix="test_names_matcher_",
                            delete=not is_windows) as f1:
        f1.write("""Vadim Markovtsev
Waren Long""")
        f1.flush()
        with NamedTemporaryFile(mode="w+", prefix="test_names_matcher_",
                                delete=not is_windows) as f2:
            f2.write("""Long Waren
vmarkovtsev|vadim
""")
            f2.flush()
            if is_windows:
                f1.close()
                f2.close()
            with patch.object(sys, "argv", ["", f1.name, f2.name]):
                try:
                    main()
                finally:
                    if is_windows:
                        os.remove(f1.name)
                        os.remove(f2.name)
            assert capsys.readouterr().out == """vmarkovtsev|vadim
Long Waren
"""



from typing import Iterable

import numpy as np
import pytest

from names_matcher.algorithm import NamesMatcher


def test_docstring():
    assignments = NamesMatcher()([["Vadim Markovtsev", "vmarkovtsev"], ["Long, Waren", "warenlg"]],
                                 [["Warren"], ["VMarkovtsev"], ["Eiso Kant"]])
    assert assignments[0].tolist() == [1, 0]
    assert assignments[1].tolist() == [1, 0.5882352941176471]


@pytest.mark.parametrize("names, result", [
    [[], (set(), "")],
    [["VMarkovtsev", "vmarkovtsev"], ({"v", "markovtsev"}, "vmarkovtsev")],
    [["markovtsev", "ricardomarkovtsev"], ({"ricardo", "markovtsev"}, "markovtsevricardo")],
    [["ricard", "ricardomarkovtsev"], ({"ricard", "omarkovtsev"}, "ricardomarkovtsev")],
    [["Vadim_MARKOVTSEV (Rebase PR)"], ({"vadim", "markovtsev"}, "vadimmarkovtsev")],
    [["MarkõvtsevVádim"], ({"vadim", "markovtsev"}, "markovtsevvadim")],
])
def test_reap_idenity(names: Iterable[str], result):
    assert NamesMatcher().reap_identity(names) == result, names


@pytest.mark.parametrize("names1, names2, result", [
    [({"a"}, "a"), ({"a"}, "a"), 0],
    [(set(), ""), (set(), ""), 1],
    [(set(), ""), ({"a"}, ""), 1],
    [({"a", "b"}, "ba"), ({"a", "c"}, "ca"), 1 / 3],
])
def test_distance(names1, names2, result):
    assert NamesMatcher.distance(names1, names2) == pytest.approx(result)


def test_progress_on(capsys):
    NamesMatcher().match_parts(*([[({"%04d" % i}, "%04d" % i) for i in range(1000)]] * 2))
    assert capsys.readouterr().err


def test_progress_off(capsys):
    NamesMatcher().match_parts(*([[({"%04d" % i}, "%04d" % i) for i in range(1000)]] * 2),
                               disable_progress=True)
    assert not capsys.readouterr().err


def test_warning():
    with pytest.warns(ResourceWarning):
        matrix = 1 - np.flip(np.eye(10000, dtype=np.float32), axis=0)
        NamesMatcher().solve_lap(matrix, len(matrix) // 2)



from importlib.machinery import SourceFileLoader
from pathlib import Path

from setuptools import find_packages, setup

project_root = Path(__file__).parent
code_root = project_root / "names_matcher"
metadata = SourceFileLoader("metadata", str(code_root / "metadata.py")).load_module()

with open(project_root / "README.md", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name=metadata.__package__.replace(".", "-"),
    description=metadata.__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=metadata.__version__,
    license="Apache-2.0",
    author="Athenian",
    author_email="vadim@athenian.co",
    url="https://github.com/athenian/names-matcher",
    packages=find_packages(exclude=["tests"]),
    keywords=["fuzzy matching"],
    install_requires=["numpy>=1.20.0",
                      "lapjv>=1.3.14,<2.0",
                      "fuzzywuzzy>=0.18.0,<2.0",
                      "unidecode>=1.0.0,<2.0",
                      "python-Levenshtein>=0.12,<2.0",
                      ],
    tests_require=[],
    package_data={
        "": ["README.md"],
        "names_matcher": ["../requirements.txt"],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: Apache Software License",
    ],
)



import re
from typing import Iterable, Sequence, Set, Tuple, Union
import warnings

from fuzzywuzzy import fuzz
from lapjv import lapjv
import numpy as np
try:
    from tqdm import tqdm
except ImportError:
    def tqdm(iterable):
        """No-op tqdm shim."""
        return iterable
from unidecode import unidecode


NormalizedIdentity = Tuple[Set[str], str]


class NamesMatcher:
    """
    Fuzzy names matching algorithm.

    Let's define an identity as a series of names belonging to the same person.

    1. Parse, normalize, and split names in each identity. The result is a set of strings per each.
    2. Define the similarity between identities as `max(ratio, token_set_ratio)`, where `ratio` \
       and `token_set_ratio` are inspired by string comparison functions from FuzzyWuzzy.
    3. Construct the distance matrix between identities in two specified lists.
    4. Solve the Linear Assignment Problem (LAP) on that matrix.

    Our LAP's solution scales up to ~1000-s of identities.

    Example:
    >>> NamesMatcher()([["Vadim Markovtsev", "vmarkovtsev"], ["Long, Waren", "warenlg"]], \
                        [["Warren"], ["VMarkovtsev"], ["Eiso Kant"]])
    (array([1, 0], dtype=int32), array([1.        , 0.58823529]))
    """

    initials_re = re.compile(r"(?<=[a-z])(?=[A-Z])|([A-Z])(?=[A-Z]+[a-z])")
    nonalphanumeric_re = re.compile(r"[^A-Za-z0-9 ]+")
    whitespace_re = re.compile(r" +")
    repetitions_re = re.compile(r"(.{2,})(.*)\1+")
    stop_words = {
        "github": ["rebase",
                   "rebasing",
                   "pr",
                   "action",
                   "script",
                   "release",
                   "unknown",
                   "root",
                   "admin",
                   "localhost",
                   "refresher",
                   "bump",
                   "lint",
                   "update",
                   "and",
                   ],
    }

    def __init__(self,
                 stop_words: Union[str, Iterable[str]] = "github",
                 ):
        """
        Initialize a new instance of NamesMatcher class.

        :param stop_words: Ignore parsed name parts that start with any string from this list. \
                           You can use regular expression syntax.
        """
        if isinstance(stop_words, str):
            stop_words = self.stop_words[stop_words]
        self._stop_words_re = re.compile("|".join(stop_words) if stop_words else "(?!.*)")

    def __call__(self,
                 names1: Iterable[Iterable[str]],
                 names2: Iterable[Iterable[str]],
                 ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Match people names from `names1` to `names2`.

        `names1` and `names2` do not have to be the same length.
        Each identity is one or several names.
        :return: numpy array with mapping indexes, so that `names2[return] ~= names1`. \
                 If there is no match found, the index is negative. \
                 numpy array with mapping confidences from 0 to 1.
        """
        return self.match_parts(
            *([self.reap_identity(n) for n in names] for names in (names1, names2)))

    def match_parts(self,
                    parts1: Sequence[NormalizedIdentity],
                    parts2: Sequence[NormalizedIdentity],
                    disable_progress: bool = False,
                    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Match parsed, normalized, split identities. You shouldn't use this function unless you \
        know what you are doing.

        `tokens1` and `tokens2` do not have to be the same length.
        Each identity is one or several parts.
        :param disable_progress: Value indicating whether to disable tqdm progress indication \
                                 for computing big distance matrices.
        :return: numpy array with mapping indexes, so that `tokens2[return] ~= tokens1`. \
                 If there is no match found, the index is negative. \
                 numpy array with mapping confidences from 0 to 1.
        """
        distances = np.ones(((offset := len(parts1)) + len(parts2),) * 2)
        distance = self.distance
        if not disable_progress and len(parts1) * len(parts2) >= 500000:
            parts1 = tqdm(parts1)
        for y, part1 in enumerate(parts1):
            for x, part2 in enumerate(parts2):
                distances[y, offset + x] = distances[offset + x, y] = distance(part1, part2)
        return self.solve_lap(distances, offset)

    @staticmethod
    def solve_lap(distances: np.ndarray, offset: int) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Solve the LAP on the given distance matrix. You shouldn't use this function unless you \
        know what you are doing.

        The distance matrix should be of this form:

                        <offset>
                  ---------|-----------------
                 | 1 1 1 1 |    distances    |  \
                 | 1 1 1 1 |     1 -> 2      |    tokens1
                 | 1 1 1 1 |                 |  /
        <offset> |---------|-----------------|
                 | distan- | 1 1 1 1 1 1 1 1 |  \
                 | ces     | 1 1 1 1 1 1 1 1 |
                 | 2 -> 1  | 1 1 1 1 1 1 1 1 |    tokens2
                 |         | 1 1 1 1 1 1 1 1 |
                 |         | 1 1 1 1 1 1 1 1 |  /
                  ---------|-----------------
                  \       / \               /
                   tokens1        tokens2
        """  # noqa: D400, D205
        if distances.shape[0] >= 10000:
            warnings.warn("Matching %d x %d names is likely a heavy LAP and will take much time." %
                          (offset, len(distances) - offset), ResourceWarning)
        row_ind, _, _ = lapjv(distances)
        assignments = row_ind[:offset] - offset
        confidences = 1 - distances[tuple(np.vstack([np.arange(offset), row_ind[:offset]]))]
        # there can be matches (1) -> (1) in case everything in (2) is at distance 1
        confidences[assignments < 0] = 1
        assignments[assignments < 0] = -1
        return assignments, confidences

    @staticmethod
    def distance(id1: NormalizedIdentity, id2: NormalizedIdentity) -> float:
        """
        FuzzyWuzzy distance between two identities.

        We take the max of `token_set_ratio()` and `ratio()`.
        """  # noqa: D403
        tokens1, concat1 = id1
        tokens2, concat2 = id2
        if not tokens1 or not tokens2:
            return 1
        assert concat1
        assert concat2

        # the following is a copy-paste from FuzzyWuzzy's token_set_ratio()
        intersection = tokens1.intersection(tokens2)
        diff1to2 = tokens1.difference(tokens2)
        diff2to1 = tokens2.difference(tokens1)

        sorted_sect = " ".join(sorted(intersection))
        sorted_1to2 = " ".join(sorted(diff1to2))
        sorted_2to1 = " ".join(sorted(diff2to1))

        combined_1to2 = sorted_sect + " " + sorted_1to2
        combined_2to1 = sorted_sect + " " + sorted_2to1

        sorted_sect = sorted_sect.strip()
        combined_1to2 = combined_1to2.strip()
        combined_2to1 = combined_2to1.strip()

        # Pick the best our of 4: the normalized concatentation and the 3 set variants
        pairwise = [fuzz.SequenceMatcher(None, *pair).ratio() for pair in (
            (concat1, concat2),
            (sorted_sect, combined_1to2),
            (sorted_sect, combined_2to1),
            (combined_1to2, combined_2to1),
        )]

        return 1 - max(pairwise)

    def reap_identity(self, names: Iterable[str]) -> NormalizedIdentity:
        """
        Parse, normalize, and split names to a set of characteristic strings.

        :return: 1. Set of normalized name atoms. \
                 2. Normalized names concatenation.
        """
        initials_re = self.initials_re
        nonalphanumeric_re = self.nonalphanumeric_re
        whitespace_re = self.whitespace_re
        stop_words_re = self._stop_words_re

        parts = set()
        amalgamation = []
        for name in names:
            normal = unidecode(name)
            concat = ""
            for variant in initials_re.split(normal):
                # break by lower -> UPPER case change and by UPPER -> UPPER+ -> lower
                # examples:
                # VadimMarkovtsev -> Vadim Markovtsev
                # VMarkovtsev -> V Markovtsev
                if not variant:
                    continue
                variant = variant.lower()
                # replace all non-alpha and non-numeric chars with a whitespace
                variant = nonalphanumeric_re.sub(" ", variant)
                # collapse sequential whitespaces
                variant = whitespace_re.sub(" ", variant)
                variant = variant.strip()
                if not variant:
                    continue
                for part in variant.split(" "):
                    # if we start with a stop word, ignore
                    if stop_words_re.match(part):
                        continue
                    concat += part
                    parts.add(part)
            amalgamation.append(concat)
        amalgamation = "".join(sorted(amalgamation))
        amalgamation = self.repetitions_re.sub(r"\1\2", amalgamation)

        # remove compounds, split them by prefix and by suffix
        while True:
            removed = []
            added = []
            for part1 in parts:
                for part2 in parts:
                    if part1 != part2:
                        if (concat := (part1 + part2)) in parts:
                            removed.append(concat)
                        if len(part1) >= 2:
                            if part2.startswith(part1):
                                removed.append(part2)
                                added.append(part2[len(part1):])
                            elif part2.endswith(part1):
                                removed.append(part2)
                                added.append(part2[:-len(part1)])
            for part in removed:
                try:
                    parts.remove(part)
                except KeyError:
                    continue
            parts.update(added)
            if not removed and not added:
                break
        return parts, amalgamation
